from datetime import datetime
from typing import Literal, Optional, Sequence

from pydantic import BaseModel, field_serializer


class Person(BaseModel):
    personid: int
    plastname: str
    pfirstname: str
    psurname: str
    ppsex: int
    isuser: bool
    pilogin: str
    pipassword: str
    caid: str
    caidname: str
    rspostid: str
    rspostidname: str
    personemail: str
    pstatus: int
    pextcode: str

    def __str__(self) -> str:
        return f"{self.plastname} {self.pfirstname} {self.psurname}"


class Persons:
    def __init__(self, persons: Sequence[Person], count: int):
        self._persons = sorted(persons, key=lambda x: x.personid)
        self.count = count

    def __getitem__(self, key: int) -> Person:
        return self._persons[key]

    def __str__(self) -> str:
        return "\n".join([str(person) for person in self._persons])
    
    def model_dump(self):
        return [model.model_dump() for model in self._persons]


class Measure(BaseModel):
    meid: Optional[int] = None
    mename: str
    medescription: Optional[str] = None
    metype: Literal["0", "1", "2"] = "1"
    mecode: Optional[str] = None
    mestatus: Optional[int] = None
    mestartdate: Optional[datetime] = None
    meenddate: Optional[datetime] = None
    meeduform: Literal["0", "1", "2"] = "1"
    mecontenttype: Optional[int] = None
    testid: Optional[str] = None
    testidname: Optional[str] = None
    ugrid: Optional[int] = None
    ugridname: Optional[str] = None
    mepasses: Optional[int] = None

    def __str__(self):
        result = f"{self.mename}\n"
        if self.mestartdate:
            result += f"Начало: {
                self.mestartdate.strftime('%d.%m.%Y %H:%M')}\n"
        if self.meenddate:
            result += f"Окончание: {
                self.meenddate.strftime('%d.%m.%Y %H:%M')}\n"
        return result

    @field_serializer("mestartdate", "meenddate")
    def format_date(self, dt: Optional[datetime]):
        if dt:
            return dt.strftime("%Y-%m-%d %H:%M:%S.000")


class Measures:
    def __init__(self, measures: Sequence[Measure], count: int):
        self._measures = sorted(measures, key=lambda x: x.meid, reverse=True)
        self.count = count

    def __getitem__(self, key: int) -> Measure:
        return self._measures[key]

    def __str__(self):
        return "\n".join([str(measure) for measure in self._measures])
    
    def model_dump(self):
        return [model.model_dump() for model in self._measures]


class Member(BaseModel):
    isaccess: bool
    meid: int
    meidname: str
    mmfinishstatus: int
    mmid: int
    personid: int
    personidname: str

    def __str__(self):
        return self.personidname


class Members:
    def __init__(self, members: Sequence[Member], count: int):
        self._members = members
        self.count = count

    def __getitem__(self, key: int) -> Member:
        return self._members[key]

    def __str__(self):
        return "\n".join([member.personidname for member in self._members])
    
    def model_dump(self):
        return [model.model_dump() for model in self._members]


class Tutor(BaseModel):
    isaccess: bool
    meid: int
    meidname: str
    mtid: int
    personid: int
    personidname: str

    def __str__(self):
        return self.personidname


class Tutors:
    def __init__(self, tutors: Sequence[Tutor], count: int):
        self._tutors = tutors
        self.count = count

    def __getitem__(self, key: int) -> Tutor:
        return self._tutors[key]

    def __str__(self):
        return "\n".join([tutor.personidname for tutor in self._tutors])
    
    def model_dump(self):
        return [model.model_dump() for model in self._tutors]


class MeasureResult(BaseModel):
    mmid: int
    mrstatusid: Optional[int] = None
    mrlastaccesstime: Optional[datetime] = None
    mrscoreinpercent: Optional[int] = None
    mrscore: Optional[float] = None
    mrprocent: Optional[int] = None
    mrduration: Optional[float] = None
    mrstarttime: Optional[datetime] = None
    mrendtime: Optional[datetime] = None
    mrattemptcount: Optional[int] = None


class MeasureResults:
    def __init__(self, measure_results: Sequence[MeasureResult], count: int):
        self._measure_results = measure_results
        self.count = count

    def __getitem__(self, key: int) -> MeasureResult:
        return self._measure_results[key]

    def model_dump(self):
        return [model.model_dump() for model in self._measure_results]


class WebinarRecord(BaseModel):
    id: str
    name: str
    period: Optional[str] = None
    duration: Optional[str] = None
    viewlink: str
    downloadlink: Optional[str] = None

    def __str__(self) -> str:
        return self.viewlink


class WebinarRecords:
    def __init__(self, webinar_records: Sequence[WebinarRecord], count: int):
        self._webinar_records = webinar_records
        self.count = count

    def __getitem__(self, key: int) -> WebinarRecord:
        return self._webinar_records[key]

    def model_dump(self):
        return [model.model_dump() for model in self._webinar_records]



from typing import Optional

from .base_api import BaseAPI
from .models import (
    Measure,
    MeasureResult,
    MeasureResults,
    Measures,
    Member,
    Members,
    Person,
    Persons,
    Tutor,
    Tutors,
    WebinarRecord,
    WebinarRecords,
)


class VirtualRoom(BaseAPI):
    def __init__(
        self,
        base_link: str,
        secret_key: str,
        app_id: str,
        service_path: str = "/service/v2",
        verify_ssl: bool = True,
    ):
        super().__init__(
            api_link=base_link + service_path,
            secret_key=secret_key,
            app_id=app_id,
            verify_ssl=verify_ssl,
        )

    async def get_persons(self, limit: int = 200, offset: int = 0) -> Optional[Persons]:
        """
        Получение информации о физических лицах
        :rtype: list[Person]
        :param limit: Количество записей (200 максимально)
        :param offset: Сдвиг страницы
        :return: Sequence of Persons and count
        """
        persons = await self._get(
            route="/persons",
            params={
                "limit": limit,
                "offset": offset,
            },
        )
        if persons:
            return Persons([Person(**person) for person in persons["data"]], persons["count"])
        else:
            return None

    async def get_person(self, person_id: int) -> Optional[Person]:
        """
        Получение информации о физическом лице
        :rtype: Person
        :param person_id: идентификатор физического лица
        :return: Person
        """
        person = await self._get(route=f"/persons/{person_id}")
        if person:
            return Person(**person)
        else:
            return None

    async def get_measures(self, limit: int = 200, offset: int = 0) -> Optional[Measures]:
        """
        Получение информации о мероприятиях
        :param limit: Количество записей (200 максимально)
        :param offset: Сдвиг страницы
        :return: Sequence of Measures and count
        """
        measures = await self._get(
            route="/measures",
            params={
                "limit": limit,
                "offset": offset,
            },
        )
        if measures:
            return Measures([Measure(**measure) for measure in measures["data"]], measures["count"])
        else:
            return None

    async def get_measure(self, measure_id: int) -> Optional[Measure]:
        """
        Получение информации о мероприятии
        :rtype: Measure
        :param measure_id: идентификатор мероприятия
        :return: Measure
        """
        measure = await self._get(route=f"/measures/{measure_id}")
        if measure:
            return Measure(**measure)
        else:
            return None

    async def get_members(self, measure_id: int, limit: int = 200, offset: int = 0) -> Optional[Members]:
        """
        Получение информации об участниках мероприятия
        :param measure_id: идентификатор мероприятия
        :param limit: Количество записей (200 максимально)
        :param offset: Сдвиг страницы
        :return: Sequence of Members and count
        """
        members = await self._get(
            route=f"/measures/{measure_id}/members",
            params={
                "limit": limit,
                "offset": offset,
            },
        )
        if members:
            return Members([Member(**member) for member in members["data"]], members["count"])
        else:
            return None

    async def get_tutors(self, measure_id: int, limit: int = 200, offset: int = 0) -> Optional[Tutors]:
        """
        Получение информации о преподавателях мероприятия
        :param measure_id: идентификатор мероприятия
        :param limit: Количество записей (200 максимально)
        :param offset: Сдвиг страницы
        :return: Sequence of tutors and count
        """
        tutors = await self._get(
            route=f"/measures/{measure_id}/tutors",
            params={
                "limit": limit,
                "offset": offset,
            },
        )
        if isinstance(tutors, dict):
            return Tutors([Tutor(**tutor) for tutor in tutors["data"]], tutors["count"])
        if isinstance(tutors, list):
            return Tutors([Tutor(**tutor) for tutor in tutors], len(tutors))
        else:
            return None

    async def get_measures_info(self):
        measures_info = await self._get(route="/measures/info")
        return measures_info

    async def get_measure_results(self, measure_id: int, limit: int = 200, offset: int = 0):
        measure_results = await self._get(
            route=f"/measures/{measure_id}/results",
            params={
                "limit": limit,
                "offset": offset,
            },
        )
        if measure_results:
            return MeasureResults(
                [MeasureResult(**measure_result) for measure_result in measure_results["data"]],
                measure_results["count"],
            )

    async def create_measure(
        self,
        measure: Measure,
    ) -> Optional[Measure]:
        """
        Добавление мероприятия
        :param measure: объект мероприятия
        :return: созданное мероприятие
        """
        created_measure = await self._post(
            route="/measures",
            data=measure.model_dump(exclude_none=True),
        )
        if created_measure:
            return Measure(**created_measure)

    async def add_tutor_to_measure_by_email(
        self,
        measure_id: int,
        tutor_email: str,
        send_notifications: bool = True,
        add_roles_by_default: bool = True,
        enable_search_by_email: bool = True,
    ):
        """
        Добавление преподавателя мероприятия по E-mail
        :param measure_id: идентификатор мероприятия
        :param tutor_email: e-mail преподавателя
        :return: id физического лица
        """
        data = {
            "sendNotifications": send_notifications,
            "addRolesByDefault": add_roles_by_default,
            "enableSearchByEmail": enable_search_by_email,
        }
        tutor = await self._post(route=f"/measures/{measure_id}/tutors/regbyemail/{tutor_email}", data=data)
        return tutor

    async def add_tutor_to_measure_by_id(
        self,
        measure_id: int,
        person_id: int,
    ) -> Optional[Tutor]:
        """
        Добавление преподавателя мероприятия по id
        :param measure_id: идентификатор мероприятия
        :param person_id: id физического лицапреподавателя
        :return: Tutor
        """
        tutor = await self._post(
            route=f"/measures/{measure_id}/tutors/{person_id}",
        )
        if tutor:
            return Tutor(**tutor)

    async def delete_measure(self, measure_id: int) -> Optional[bool]:
        """
        Удаление мероприятия
        :param measure_id: идентификатор мероприятия
        :return: True if success
        """
        measure = await self._delete(route=f"/measures/{measure_id}")
        if measure:
            return True
        else:
            return False

    async def get_measure_guest_link(
        self,
        measure_id: int,
    ) -> Optional[str]:
        """
        Получение ссылки гостевого входа на вебинар
        :param measure_id: идентификатор мероприятия
        :return: Ссылка для регистрации на вебинар
        """
        guest_link = await self._get(route=f"/measures/{measure_id}/webinarAnonymousLink")
        if guest_link:
            return guest_link
        else:
            return None

    async def get_webinar_records(self, measure_id: int, limit: int = 200, offset: int = 0) -> Optional[WebinarRecords]:
        """
        Получение ссылок на записи вебинара
        """
        webinar_records = await self._get(
            route=f"/measures/{measure_id}/webinarRecords",
            params={
                "limit": limit,
                "offset": offset,
            },
        )
        if webinar_records:
            return WebinarRecords(
                [WebinarRecord(**webinar_record) for webinar_record in webinar_records["data"]],
                webinar_records["count"],
            )
        else:
            return None



import logging
from hashlib import md5
from json import dumps
from typing import Optional
from urllib.parse import parse_qsl, urlencode, urlparse, urlunparse

from aiohttp import ClientConnectionError, ClientSession


class BaseAPI:
    def __init__(
        self,
        api_link: str,
        secret_key: str,
        app_id: str,
        verify_ssl: bool = False,
    ):
        self._link = api_link
        self._verify_ssl = verify_ssl

        self._secret_key = secret_key
        self._app_id = app_id

        self._api_link = api_link
        self._base_params = {
            "appid": self._app_id,
            "secretkey": self._secret_key,
        }

        self._headers = {
            "Accept": "*/*",
        }

    async def _prepare_params(self, route: str, params: Optional[dict] = None):
        if params is None:
            params = {}

        url_parts = list(urlparse(self._api_link + route))
        query = dict(parse_qsl(url_parts[4]))

        params.update(self._base_params)
        query.update(params)

        url_parts[4] = urlencode(query)

        legacy_url = urlunparse(url_parts)
        md5_hash = md5(legacy_url.encode())
        sign = md5_hash.hexdigest().upper()

        encrypted_params = {"appid": self._app_id, "sign": sign}
        params.update(encrypted_params)
        params.pop("secretkey")
        return params

    async def _get(self, route: str, params: Optional[dict] = None):
        """
        Send get request to host
        :param params: request params
        :param route: request link
        :return: json object from host
        """
        params = await self._prepare_params(route, params)
        logging.info(f"GET {self._link}{route} {params=}")
        try:
            async with ClientSession(headers=self._headers) as session:
                async with session.get(
                    url=f"{self._link}{route}",
                    params=params,
                    verify_ssl=self._verify_ssl,
                ) as get:
                    if get.ok:
                        if get.content_type == "text/plain":
                            answer = await get.text()
                            logging.info(
                                f"GET {get.status} {
                                self._link}{route} {get.content_type} {answer=}"
                            )
                            return answer
                        answer = await get.json()
                        logging.info(
                            f"GET {get.status} {
                            self._link}{route} {get.content_type} {answer=}"
                        )
                        if "Content-Range" in get.headers:
                            logging.info(f"GET Content-Range: {get.headers["Content-Range"]}")
                            count = int(get.headers["Content-Range"].split("/")[-1])
                            return {"data": answer, "count": count}
                        else:
                            return answer
                    else:
                        logging.warning(
                            f"GET {get.status} {
                            self._link}{route}"
                        )
                        error = await get.json()
                        raise ClientConnectionError(error)
        except ClientConnectionError as e:
            logging.warning(f"Api error: {self._link}{route} {e}")
        except Exception as e:
            logging.warning(f"Api is unreachable: {e}")

    async def _post(
        self,
        route: str,
        params: Optional[dict] = None,
        data: Optional[dict] = None,
    ) -> Optional[dict | str]:
        """
        Send post request to host
        :param params: request params
        :param data: request data
        :param route: request link
        :return: json object from host
        """
        params = await self._prepare_params(route, params)
        logging.info(f"POST {self._link}{route} {params=} {data=}")
        headers = {"Content-Type": "application/json"}
        try:
            async with ClientSession(headers=headers) as session:
                async with session.post(
                    f"{self._link}{route}",
                    params=params,
                    data=dumps(data, indent=4),
                    verify_ssl=self._verify_ssl,
                ) as post:
                    if post.ok:
                        if post.content_type == "text/plain":
                            answer = await post.text()
                            logging.info(
                                f"POST {post.status} {
                                self._link}{route} {post.content_type} {answer=}"
                            )
                            return answer
                        answer = await post.json()
                        logging.info(
                            f"POST {post.status} {
                            self._link}{route} {post.content_type} {answer=}"
                        )
                        return answer
                    else:
                        logging.warning(
                            f"POST {post.status} {
                            self._link}{route}"
                        )
                        error = await post.json()
                        raise ClientConnectionError(error)
        except ClientConnectionError as e:
            logging.warning(f"Api error: {self._link}{route} {e}")
        except Exception as e:
            logging.warning(f"Api is unreachable: {e}")

    async def _delete(self, route: str, params: Optional[dict] = None) -> Optional[int]:
        params = await self._prepare_params(route, params)
        logging.info(f"DELETE {self._link}{route} {params=}")
        try:
            async with ClientSession(headers=self._headers) as session:
                async with session.delete(
                    url=f"{self._link}{route}",
                    params=params,
                    verify_ssl=self._verify_ssl,
                ) as delete:
                    if delete.ok:
                        logging.info(
                            f"DELETE {delete.status} {
                            self._link}{route}"
                        )
                        return delete.status
                    else:
                        logging.warning(
                            f"DELETE {delete.status} {
                            self._link}{route}"
                        )
                        error = await delete.json()
                        raise ClientConnectionError(error)
        except ClientConnectionError as e:
            logging.warning(f"Api error: {self._link}{route} {e}")
        except Exception as e:
            logging.warning(f"Api is unreachable: {e}")
